# -*- coding: utf-8 -*-

import socket
import ssl
import threading


class IsNotCallableException(Exception):
    pass


class SocketServer(object):


    def __init__(self):
        
        self.socket = None

        # @params: local_address, local_port
        self.onBindEvent = None

        # @params: void
        self.onWaitConnectionEvent = None

        # @params: remote_address, remote_port
        self.onClientConnectEvent = None

        # @params: remote_address, remote_port
        self.onClientCloseEvent = None

        # @params: connection, remote_address, remote_port, data
        self.onClientReceiveEvent = None


    def close(self):

        self.socket.close()


    def bindOne(self, local_address, local_port):

        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.bind((str(local_address), int(local_port)))
        self.socket.listen(0)

        # User event
        if(callable(self.onBindEvent)):
            self.onBindEvent(local_address, local_port)

        # User event
        if(callable(self.onWaitConnectionEvent)):
            self.onWaitConnectionEvent()

        # Wait for new connection
        conn, client = self.socket_handler.accept()

        # Send to event
        self.onClientConnect(
            connection     = conn,
            remote_address = client[0],
            remote_port    = client[1]
        )
        
        self.close()


    def bind(self, local_address, local_port):

        self.socket_handler = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket_handler.bind((str(local_address), int(local_port)))
        self.socket_handler.listen(1)

        # User event
        if(callable(self.onBindEvent)):
            self.onBindEvent(local_address, local_port)

        while(True):

            # User event
            if(callable(self.onWaitConnectionEvent)):
                self.onWaitConnectionEvent()

            # Wait for new connection
            conn, client = self.socket_handler.accept()
            
            # Send to thread
            threadHandler = threading.Thread(
                target = self.onClientConnect,
                args   = (conn, str(client[0]), int(client[1]))
            )

            # End thread when end main process
            threadHandler.setDaemon(True)
            threadHandler.start()


    def onClientConnect(self, connection, remote_address, remote_port):

        # User event
        if(callable(self.onClientConnectEvent)):
            self.onClientConnectEvent(remote_address, remote_port)

        with connection:

            while(True):

                data = connection.recv(1024)

                if not data:
                    break

                if(callable(self.onClientReceiveEvent)):
                    self.onClientReceiveEvent(
                        connection,
                        remote_address,
                        remote_port,
                        data
                    )

                # connection.sendall(data)

        # User event
        if(callable(self.onClientCloseEvent)):
            self.onClientCloseEvent(remote_address, remote_port)


    def setOnBind(self, function):

        # Is callable?
        if(not callable(function)):
            raise IsNotCallableException()

        # Hook event
        self.onBindEvent = function


    def setOnWaitConnection(self, function):

        # Is callable?
        if(not callable(function)):
            raise IsNotCallableException()

        # Hook event
        self.onWaitConnectionEvent = function


    def setOnClientConnect(self, function):

        # Is callable?
        if(not callable(function)):
            raise IsNotCallableException()

        # Hook event
        self.onClientConnectEvent = function


    def setOnClientClose(self, function):

        # Is callable?
        if(not callable(function)):
            raise IsNotCallableException()

        # Hook event
        self.onClientCloseEvent = function


    def setOnClientReceive(self, function):

        # Is callable?
        if(not callable(function)):
            raise IsNotCallableException()

        # Hook event
        self.onClientReceiveEvent = function



class SocketClient(object):


    def __init__(self):

        # Socket handler
        self.socket = None

        # Line control
        self.line_part = b''

        # Satus control
        self.is_alive = False

        # @params: hostname, port, use_ssl
        self.onConnecting = None

        # @params: hostname, port, use_ssl
        self.onConnect = None

        # @params: void
        self.onClose = None

        # @params: binary_data
        self.onSent = None

        # @params: binary_data
        self.onReceive = None

        # @params: binary_line
        self.onReceiveLine = None


    def close(self):

        # Close once time
        if(not self.is_alive):
            return

        if(self.socket):
            try:
                self.socket.close()
            except Exception as ignored:
                pass

        self.is_alive = False

        # User event
        if(callable(self.onClose)):
            self.onClose()


    def connect(self, hostname, port, use_ssl=False):

        # Reset buffer line
        self.line_part = b''

        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.settimeout(40)

        # User event
        if(callable(self.onConnecting)):
            self.onConnecting(str(hostname), int(port), bool(use_ssl))

        # Use ssl?
        if(bool(use_ssl)):
            self.socket = ssl.create_default_context().wrap_socket(
                self.socket,
                server_hostname=str(hostname)
            )

        self.socket.connect((str(hostname), int(port)))

        self.is_alive = True

        # User event
        if(callable(self.onConnect)):
            self.onConnect(str(hostname), int(port), bool(use_ssl))

        while(self.is_alive):

            try:
                binary_data = self.socket.recv(2048)

                if(not binary_data):
                    break

            except Exception as ignored:
                break

            # User event
            if(callable(self.onReceive)):
                self.onReceive(binary_data)

            # Buffered by line?
            if(not callable(self.onReceiveLine)):
                continue

            # Append data to buffer
            self.line_part += binary_data

            # Have a new line?
            if(not b'\n' in self.line_part):
                continue

            # Split all lines by receive and last data
            lines = self.line_part.split(b'\n')

            # Last part for continue a new line
            self.line_part = lines.pop()

            # Call user event for each line
            for line in lines:

                # Striped for drop \r
                self.onReceiveLine(line.strip())

        self.close()


    def send(self, binary_data):

        if(self.socket):
            try:
                self.socket.sendall(binary_data)

            except BrokenPipeError as e:

                # Socket is closed
                self.close()

            except Exception as e:

                # Unhandled exception
                raise e

        # User event
        if(callable(self.onSent)):
            self.onSent(binary_data)


    def setOnConnecting(self, function):

        # Is callable?
        if(not callable(function)):
            raise IsNotCallableException()

        # Hook event
        self.onConnecting = function


    def setOnConnect(self, function):

        # Is callable?
        if(not callable(function)):
            raise IsNotCallableException()

        # Hook event
        self.onConnect = function


    def setOnClose(self, function):

        # Is callable?
        if(not callable(function)):
            raise IsNotCallableException()

        # Hook event
        self.onClose = function


    def setOnSent(self, function):

        # Is callable?
        if(not callable(function)):
            raise IsNotCallableException()

        # Hook event
        self.onSent = function


    def setOnReceive(self, function):

        # Is callable?
        if(not callable(function)):
            raise IsNotCallableException()

        # Hook event
        self.onReceive = function


    def setOnReceiveLine(self, function):

        # Is callable?
        if(not callable(function)):
            raise IsNotCallableException()

        # Hook event
        self.onReceiveLine = function




# -*- coding: utf-8 -*-


class BruteDictionaryIter(object):


    def __init__(
        self,
        maxLength  = 3,
        characters = None,
        lowercase  = None,
        uppercase  = None,
        numbers    = None,
        specials   = None
    ):
        if(characters):
            self.characters = characters

        else:
            self.characters = ''
            
            if(lowercase):
                self.characters += 'abcdefghijklmnopqrstuvwxyz'

            if(uppercase):
                self.characters += 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'

            if(numbers):
                self.characters += '0123456789'

            if(specials):
                self.characters += '\\/\'"@#$%&/()[]{}=?¿!¡+-*_.:,;^'

        self.maxLength = maxLength
        self.currentIndexStack = 0

        # Stack of indexes of characters
        self.stack = []
        for n in range(0, maxLength):
            self.stack.append(None)

        # Statics
        self.totalWords       = (len(self.characters) ** maxLength)
        self.currentWordIndex = 0

    
    def __iter__(self):

        return self


    def __next__(self):

        # For user statics only
        self.currentWordIndex += 1

        if(
            (not self.characters) or
            (len(self.stack) == 0)
        ):
            raise StopIteration

        if(self.stack[self.currentIndexStack] is None):
            
            # First value
            self.stack[self.currentIndexStack] = 0
        else:
            # Next value
            self.stack[self.currentIndexStack] += 1

        # Overflow?
        if(self.stack[self.currentIndexStack] > (len(self.characters) - 1)):

            # Reset counter
            self.stack[self.currentIndexStack] = 0

            # Sum previous index
            prevIndex = self.currentIndexStack - 1
            while(True):

                # No more previous index?, need append a unity
                if(prevIndex < 0):

                    # Next index
                    self.currentIndexStack += 1

                    if(self.currentIndexStack > (len(self.stack) - 1)):

                        # No more indexes on stack
                        raise StopIteration

                    # First value of next stack
                    self.stack[self.currentIndexStack] = 0

                    break

                else:
                    # Next value of previous index
                    self.stack[prevIndex] += 1

                    # Overflow?
                    if(self.stack[prevIndex] > (len(self.characters) - 1)):

                        self.stack[prevIndex] = 0

                        # Try next with previous index
                        prevIndex -= 1

                    else:
                        # Normal operation
                        break
            

        # Make the final string
        characters = ''

        for indexWord in self.stack:

            # Check for right padding using None
            if(indexWord is None):
                continue

            characters += self.characters[indexWord]

        return characters



# _*_ coding:utf-8 _*_

import re
from urllib import parse
from urllib.parse import urlencode

from pentest.net.socket import SocketClient


class HttpClient(object):
    

    def __init__(self):

        # Local memory storage for cookies
        self.cookies  = {}

        # Last url for sent the referer header
        self.last_url = None

        # Socket helper handler
        self.socket_client = SocketClient()

        # Handle events
        self.socket_client.setOnReceive(self.onReceive)
        self.socket_client.setOnConnect(self.onConnect)

        # Current request context
        self.current_context = {
            'hostname'       : None,
            'port'           : None,
            'use-ssl'        : None,
            'url'            : None,
            'uri'            : None,
            'custom-headers' : None,
            'post-data'      : None,
            'raw-request'    : None,
            'raw-response'   : None
        }


    def clearContext(self):

        self.cookies  = {}
        self.last_url = None


    def request(self, url, custom_headers=None, post_data=None):
        
        # to bytes array
        if(not isinstance(url, bytes)):
            url = str(url).encode('utf-8', 'ignore')

        if(post_data):
            
            if(isinstance(post_data, dict)):
                post_data = urlencode(post_data).encode('utf-8', 'ignore')

            if(isinstance(post_data, str)):
                post_data = post_data.encode('utf-8', 'ignore')

        # if(custom_headers):
        #     # Join custom headers
        #     headers.update(custom_headers)

        # Parse URL
        url_parsed = parse.urlparse(url)
        url_data   = {
            'original' : url,
            'path'     : url_parsed.path,
            'host'     : url_parsed.netloc,
            'port'     : url_parsed.port,
            'scheme'   : url_parsed.scheme,
            'query'    : url_parsed.query,
            'uri'      : url_parsed.path + ((b'?' + url_parsed.query) if url_parsed.query else b'')
        }

        if url_data['path'] == b'':
            url_data['path'] = b'/'

        if(not url_data['port']):
            url_data['port'] = 443 if (url_data['scheme'] == b'https') else 80
        
        # Strip the custom port from the hostname
        if(b':' in url_data['host']):
            url_data['host'] = url_data['host'].split(b':')[0]

        # Copy arguments to current context
        self.current_context['hostname']       = url_data['host'].decode()
        self.current_context['port']           = int(url_data['port'])
        self.current_context['use-ssl']        = bool(url_data['scheme'] == b'https')
        self.current_context['url']            = str(url)
        self.current_context['uri']            = url_data['uri'].decode()
        self.current_context['custom-headers'] = custom_headers
        self.current_context['post-data']      = post_data
        self.current_context['raw-response']   = b''

        # Connect to server
        self.socket_client.connect(
            hostname = self.current_context['hostname'],
            port     = self.current_context['port'],
            use_ssl  = self.current_context['use-ssl']
        ) # Locked for connection

        # The socket is closed in this point

        # Save the cookies for next request
        matches = re.findall(
            br'set\-cookie:\s*(.*?);',
            self.current_context['raw-response'],
            re.IGNORECASE | re.MULTILINE
        )

        if len(matches) > 0:
            for cookie in matches:

                cookie = re.search(
                    rb'(.*?)=(.*)',
                    cookie,
                    re.IGNORECASE | re.MULTILINE
                )
                
                var = cookie.group(1).strip()
                val = cookie.group(2).strip()

                if val:
                    self.cookies[var] = val

                else:
                    # Delete cookie if exist
                    if var in self.cookies.keys():
                        self.cookies.pop(var, None)

        # Get the HTTP status code
        status_code = 0
        matches     = re.search(
            br'HTTP\/\d\.\d (\d+) ',
            self.current_context['raw-response'],
            re.IGNORECASE | re.MULTILINE
        )
        if(matches):
            status_code = int(matches.group(1))

        # Split the header and body
        body    = self.current_context['raw-response'].split(b'\r\n\r\n')
        headers = body.pop(0).strip()
        body    = b'\r\n'.join(body)

        # Free memory
        self.current_context['raw-response'] = b''

        # Split headers
        if(b'\r\n' in headers):
            tmp = {}
            for item in headers.split(b'\r\n'):
                value    = item.split(b':')
                key      = value.pop(0).strip()
                value    = b':'.join(value).strip()
                tmp[key] = value
            headers = tmp

        # Parse the raw body

        # TODO: Add better support
        # Transfer-Encoding: chunked  [OK]
        # Transfer-Encoding: compress
        # Transfer-Encoding: deflate
        # Transfer-Encoding: gzip
        # Transfer-Encoding: identity

        if(b'Transfer-Encoding' in headers):
        
            # Chunked response body?
            # https://tools.ietf.org/rfc/rfc7230.txt
            if(headers[b'Transfer-Encoding'] == b'chunked'):

                # Length in hex, convert to int base 16
                bytes_length = b''
                buffer_body  = b''
                body         = b'\r\n' + body # Padding left for bytes length

                while(True):

                    # [\r\n] [length] [\r\n] [part]

                    groups       = re.search(br'\r\n(\w+?)\r\n', body)
                    bytes_length = groups.group(1)
                    bytes_length = int(bytes_length, 16) # Convert length value
                    body         = body[len(groups.group(0)):]# Strip the length string

                    if(bytes_length == 0):
                        break
                    
                    buffer_body      += body[:bytes_length]
                    body             =  body[bytes_length:]

                # Transfer value
                body = buffer_body

                # Free memory
                buffer_body = None

        # Return the result
        return {
            'status-code'      : status_code,
            'response-content' : body,
            'response-headers' : headers,
            'request-content'  : self.current_context['raw-request']
        }


    def onConnect(self, hostname, port, use_ssl):
        
        # Get the cookies
        cookies = []

        for key, value in self.cookies.items():
            cookies.append(key + b'=' + value)

        cookies = b'; '.join(cookies)

        # Begin of packet to send
        self.current_context['raw-request'] = b'\r\n'.join([
            (b'POST ' if self.current_context['post-data'] else b'GET ') + self.current_context['uri'].encode() + b' HTTP/1.1',
            b'Host: ' + self.current_context['hostname'].encode(),
            b'User-Agent: Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:65.0) Gecko/20100101 Firefox/54.0',
            b'Accept: text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
            b'Accept-Language: en-US',
            b'Referer: ' + (self.last_url.encode() if self.last_url else self.current_context['url'].encode()),
            b'Cookie: ' + cookies,
            b'Connection: close'
        ]) + b'\r\n'

        # Parse custom headers
        if(self.current_context['custom-headers']):
            custom_headers = []

            for key, value in self.current_context['custom-headers'].items():
                custom_headers.append(key.encode() + b': ' + value.encode('utf-8', errors='ignored'))

            custom_headers = b'\r\n'.join(custom_headers)
            self.current_context['raw-request'] += custom_headers + b'\r\n'

        # Send post data?
        if(self.current_context['post-data']):
            self.current_context['raw-request'] += b'\r\n'.join([
                b'Content-Type: application/x-www-form-urlencoded',
                b'Content-Length: ' + str(len(self.current_context['post-data'])).encode(),
                b'',
                self.current_context['post-data']
            ])
        else:
            self.current_context['raw-request'] += b'\r\n'

        # Save the url for sent into next request
        self.last_url = self.current_context['url']

        # Send binary packet to server
        self.socket_client.send(self.current_context['raw-request'])


    def onReceive(self, binary_data):
        
        # Append to buffer
        self.current_context['raw-response'] += binary_data



# -*- coding: utf-8 -*-


class Inputs(object):


    def __init__(self):

        self.strings = {
            'wrong-number' : 'That\'s not an number. Try again.',
            'wrong-yn'     : 'Please, respond with yes (y) or not (n). Try again.'
        }


    def get(self, prompt_message, type_obj=str, default=None):

        while(True):

            content = input(prompt_message).strip()

            if(content == ''):
                content = default

            if(type_obj is int):

                if(
                    (not str(content).isnumeric()) or
                    (int(content) < 0)
                ):
                    print(self.strings['wrong-number'])
                    continue

                return int(content)
                
            elif(type_obj == bool):

                if(
                    (str(content).lower() == 'y') or
                    (content is True)
                ):
                    return True

                elif(
                    (str(content).lower() == 'n') or
                    (content is False)
                ):
                    return False

                else:
                    print(self.strings['wrong-yn'])
                    continue

            else:
                return content